const builtin = @import("builtin");
const std = @import("std");
const Clipboard = @import("clipboard.zig");

// Define root.log to override the std implementation
// pub fn log(
//     comptime level: std.log.Level,
//     comptime scope: @TypeOf(.EnumLiteral),
//     comptime format: []const u8,
//     args: anytype,
// ) void {
//     // Ignore all non-error logging from sources other than
//     // .my_project, .nice_library and .default
//     const scope_prefix = "(" ++ switch (scope) {
//         .my_project, .nice_library, .default => @tagName(scope),
//         else => if (@enumToInt(level) <= @enumToInt(std.log.Level.err))
//             @tagName(scope)
//         else
//             return,
//     } ++ "): ";
//
//     const prefix = "[" ++ level.asText() ++ "] " ++ scope_prefix;
//
//     // Print the message a file
//     const lfile = std.fs.openFileAbsolute("d:\\yo", std.fs.File.OpenFlags{
//         .read = false,
//         .write = true,
//     }) catch return;
//     lfile.seekFromEnd(0) catch return;
//     defer lfile.close();
//
//     // Print the message to stderr, silently ignoring any errors
//     std.debug.getStderrMutex().lock();
//     defer std.debug.getStderrMutex().unlock();
//     const stderr = std.io.getStdErr().writer();
//     nosuspend lfile.writer().print(prefix ++ format ++ "\n", args) catch return;
//     nosuspend stderr.print(prefix ++ format ++ "\n", args) catch return;
// }

pub fn main() !u8 {
    var watch = false;
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();
    var args = std.process.args();
    defer args.deinit();
    while (args.next(allocator)) |arg_or_err| {
        const arg = try arg_or_err;
        defer allocator.free(arg);
        if (std.mem.eql(u8, arg, "-w")) {
            watch = true;
            break;
        }
    }
    var clip = try Clipboard.init(allocator);
    defer clip.deinit();
    if (watch) {
        while (clip.download()) |text| {
            std.log.debug("Got remote clipboard: {s}", .{text});
            addToClipboardNoError(allocator, text);
            allocator.free(text);
        }
    } else {
        if (clip.download()) |text| {
            std.log.debug("Got remote clipboard: {s}", .{text});
            addToClipboardNoError(allocator, text);
            allocator.free(text);
        }
    }

    return 0;
}
fn addToClipboardNoError(allocator: std.mem.Allocator, data: []const u8) void {
    addToClipboard(allocator, data) catch |e| {
        std.log.err("Could not add data to clipboard: {}", .{e});
        if (@errorReturnTrace()) |trace| {
            std.debug.dumpStackTrace(trace.*);
        }
    };
}

fn addToClipboard(allocator: std.mem.Allocator, data: []const u8) !void {
    // We're going to cheat here and just run xclip
    const result = os: {
        if (builtin.os.tag == .linux) {
            break :os try execLinux(allocator, data);
        } else if (builtin.os.tag == .windows) {
            break :os try execWindows(allocator, data);
        } else {
            return error.OsUnsupported;
        }
    };

    std.log.debug("exec done", .{});
    try std.io.getStdErr().writer().writeAll(result.stderr);
    switch (result.term) {
        .Exited => |code| if (code != 0) return error.NonZeroExit,
        .Signal => return error.FailedWithSignal,
        .Stopped => return error.WasStopped,
        .Unknown => return error.Failed,
    }
}

fn execLinux(allocator: std.mem.Allocator, data: []const u8) !std.ChildProcess.ExecResult {
    // we need to do this (maybe) because there is no real way to close stdin cleanly, which is what will
    // kill xclip.
    //
    // Also, it seems xclip doesn't close stdout, which triggers ChildProcess.exec to know
    // that the process is done. Redirecting stdout/stderr as below will get the job
    // done, though
    const escaped_data = try escape(allocator, data, "'", "'\"", "\"'");
    defer allocator.free(escaped_data);
    const xclip_cmd = try std.fmt.allocPrint(allocator, "echo -n '{s}'| xclip -selection c >/dev/null 2>&1", .{escaped_data});
    defer allocator.free(xclip_cmd);
    std.log.debug("cmd: {s}", .{xclip_cmd});

    return std.ChildProcess.exec(.{
        .allocator = allocator,
        .argv = &[_][]const u8{
            "/usr/bin/env",
            "sh",
            "-c",
            xclip_cmd,
        },
    });
}

fn execWindows(allocator: std.mem.Allocator, data: []const u8) !std.ChildProcess.ExecResult {
    // we need to do this (maybe) because there is no real way to close stdin cleanly, which is what will
    // kill clip.exe
    // TODO: Create minimal repro and submit zig stdlib bug
    const escaped_data = try escape(allocator, data, "'", "'", "");
    defer allocator.free(escaped_data);
    const cmd = try std.fmt.allocPrint(allocator, "write-output '{s}'| set-clipboard", .{escaped_data});
    defer allocator.free(cmd);
    std.log.debug("cmd: {s}", .{cmd});

    return std.ChildProcess.exec(.{
        .allocator = allocator,
        .argv = &[_][]const u8{
            "powershell",
            "-Command",
            cmd,
        },
    });
}

fn escape(allocator: std.mem.Allocator, data: []const u8, danger_chars: []const u8, prefix: []const u8, suffix: []const u8) ![]const u8 {
    var escaped = try std.ArrayList(u8).initCapacity(allocator, data.len);
    defer escaped.deinit();
    for (data) |c| {
        var needs_escape = false;
        for (danger_chars) |dc| {
            if (c == dc) {
                needs_escape = true;
                break;
            }
        }
        if (needs_escape) {
            try escaped.appendSlice(prefix);
            try escaped.append(c);
            try escaped.appendSlice(suffix);
        } else {
            try escaped.append(c);
        }
    }
    return escaped.toOwnedSlice();
}
